"""
a collection of argparse `type_*` and 'add_arguments_*' functions

This is a collection of functions can be used with argparser add_argumnet
and parser.

Eg.::

    # this will validate the `-x` value as HH:MM to datetime.time
    parser.add_argument("-x", type=type_hhmm)

Or::

    # this will add few arguments to the parser
    add_arguments_rexec(parser)

"""

from __future__ import annotations

import argparse
import contextlib
import datetime
import logging
from pathlib import Path
from typing import Any, Sequence

from .shared import ArgumentTypeBase, LuxosParserBase


class type_ipaddress(ArgumentTypeBase):
    """
    Validate a type as an ip addresses.

    Raises:
        argparse.ArgumentTypeError: on an invalid input.

    Returns:
        tuple[str, None | int] or None

    Example:
        file.py::

            parser.add_argument("-x", type=type_ipaddress)
            options = parser.parse_args()
            ...

            assert options.x == ("host", 9999)


        shell::

            file.py -x host:9999
    """

    def validate(self, txt) -> None | tuple[str, None | int]:
        from luxos import ips

        if txt is None:
            return None
        try:
            result = ips.parse_expr(txt) or ("", "", None)
            if result[1]:
                raise argparse.ArgumentTypeError("cannot use a range as expression")
            return (result[0], result[2])
        except ips.AddressParsingError as exc:
            raise argparse.ArgumentTypeError(f"failed to parse {txt=}: {exc.args[0]}")


def type_range(txt: str) -> Sequence[tuple[str, int | None]]:
    """
    Validate a range of ip addresses.

    Raises:
        argparse.ArgumentTypeError: on an invalid input.

    Returns:
        Sequence[tuple[str, int | None]]

    Example:
        file.py::

            parser.add_argument("-x", type=type_range)
            options = parser.parse_args()
            ...

            assert options.x == [
                ("127.0.0.1", 9999),
                ("127.0.0.2", 9999),
                ("127.0.0.3", 9999),
            ]


        shell::

            file.py -x 127.0.0.1:9999:127.0.0.3

        Alternatively you can pass a **@filename** to read data from a csv file
    """
    from luxos.ips import (
        DataParsingError,
        iter_ip_ranges,
        load_ips_from_csv,
        load_ips_from_yaml,
    )

    path = None
    if txt.startswith("@") and not (path := Path(txt[1:])).exists():
        raise argparse.ArgumentTypeError(f"file not found {path}")
    if Path(txt).exists():
        path = Path(txt)

    if path:
        with contextlib.suppress(RuntimeError, DataParsingError):
            return load_ips_from_yaml(path, None)

    try:
        if path:
            return load_ips_from_csv(path, None)
        return list(iter_ip_ranges(txt))
    except RuntimeError as exc:
        raise argparse.ArgumentTypeError(f"conversion failed '{txt}': {exc.args[0]}")
    except Exception as exc:
        raise argparse.ArgumentTypeError(
            f"unknown conversion failed for '{txt}'"
        ) from exc


class type_hhmm(ArgumentTypeBase):
    """
    Validate a type as a datetime.time in HH:MM format

    Raises:
        argparse.ArgumentTypeError: on an invalid input.

    Returns:
        datetime.time or None

    Example:
        file.py::

            parser.add_argument("-x", type=type_hhmm)
            options = parser.parse_args()
            ...

            assert options.x == datetime.time(12, 13)


        shell::

            file.py -x 12:13
    """

    def validate(self, txt) -> None | datetime.time:
        if not txt:
            return None
        with contextlib.suppress(ValueError, TypeError):
            hh, _, mm = txt.partition(":")
            hh1 = int(hh)
            mm1 = int(mm)
            return datetime.time(hh1, mm1)
        raise argparse.ArgumentTypeError(f"failed conversion into HH:MM for '{txt}'")


class type_database(ArgumentTypeBase):
    """
    Validate a type as a database string (sqlalchemy)

    Raises:
        argparse.ArgumentTypeError: on an invalid input.

    Returns:
        datetime.time or None

    Example:
        file.py::

            parser.add_argument("-x", type=type_database)
            options = parser.parse_args()
            ...

            assert options.x == sqlalchemy.Engine


        shell::

            file.py -x sqlite:///foobar.db
            file.py -x postgresql+psycopg2://<user>:<password>@<host>/<db>
    """

    def validate(self, txt) -> Any:
        if not txt:
            return None

        from sqlalchemy import create_engine
        from sqlalchemy.engine.url import make_url
        from sqlalchemy.exc import ArgumentError

        with contextlib.suppress(ArgumentError):
            url = make_url(txt)
            return create_engine(url)
        raise argparse.ArgumentTypeError(
            f"wrong sqlalchemy url format '{txt}' (eg. sqlite:///filename.db)"
        )


def add_arguments_rexec(parser: LuxosParserBase):
    """adds the rexec timing for timeout/retries/delays

    Ex.

    def add_arguments(parser):
        cli.flags.add_arguments_rexec(parser)

    def process_args(args):
        asyncops.TIMEOUT = args.timeout
        asyncops.RETRIES = args.retries
        asyncops.RETRIES_DELAY = args.retries_delay
        return args
    """
    from ..asyncops import RETRIES, RETRIES_DELAY, TIMEOUT

    group = parser.add_argument_group(
        "Remote execution", "rexec remote execution limits/timeouts"
    )
    group.add_argument(
        "--timeout", type=float, default=TIMEOUT, help="Timeout for each command"
    )
    group.add_argument(
        "--retries",
        type=int,
        default=RETRIES,
        help="Maximum number of retries for each command",
    )
    group.add_argument(
        "--retries-delay",
        type=float,
        default=RETRIES_DELAY,
        help="Delay in s between retries",
    )

    def callback(args: argparse.Namespace):
        from .. import asyncops, syncops

        asyncops.TIMEOUT = syncops.TIMEOUT = args.timeout
        asyncops.RETRIES = syncops.RETRIES = args.retries
        asyncops.RETRIES_DELAY = syncops.RETRIES_DELAY = args.retries_delay

    parser.callbacks.append(callback)


def add_arguments_config(parser: LuxosParserBase):
    # find the CONFIGPATH, from the module then from luxos.cli.v*
    default = None
    for module in reversed(parser.modules):
        if not hasattr(module, "CONFIGPATH"):
            continue
        default = Path.cwd() / Path(module.CONFIGPATH).expanduser()
        break
    if default and default.is_relative_to(Path.cwd()):
        default = default.relative_to(Path.cwd())

    parser.add_argument(
        "-c",
        "--config",
        default=default,
        type=Path,
        help="path to a config file",
    )

    def callback(args: argparse.Namespace):
        args.config = args.config.absolute() if args.config else 1

    parser.callbacks.append(callback)


def add_arguments_logging(parser: LuxosParserBase):
    group = parser.add_argument_group("Logging", "Logging related options")
    group.add_argument("-v", "--verbose", action="count", help="report verbose logging")
    group.add_argument("-q", "--quiet", action="count", help="report quiet logging")

    def setup_logging(config: dict[str, Any], count: int) -> None:
        levelmap = [
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
        ]
        n = len(levelmap)

        # awlays start from info level
        level = logging.INFO

        # we can set the default start log level in LOGGING_CONFIG
        if config.get("level", None) is not None:
            level = config["level"]

        # we control if we go verbose or quite here
        index = levelmap.index(level) + count
        config["level"] = levelmap[max(min(index, n - 1), 0)]
        logging.basicConfig(**config)

    def callback(args: argparse.Namespace):
        config = {}
        for module in reversed(args.modules):
            if not hasattr(module, "LOGGING_CONFIG"):
                continue
            config.update(module.LOGGING_CONFIG)

        count = (args.verbose or 0) - (args.quiet or 0)
        setup_logging(config, count)
        delattr(args, "verbose")
        delattr(args, "quiet")

    parser.callbacks.append(callback)


def add_arguments_database(parser: LuxosParserBase):
    """
    takes a string on a command line and retunr a sa engine.

    This is mean to be used as type in add_arguments

    .. parse_args
    args.engine = firmware.cli.process_engine(args.engine)

    The text format can be:
    - a string that will resolve into a Path
    - any string for sqlalchemy.engine.url.make_url

    In the first case a sqlite engine pointing to text Path will be returned,
    in the second a regular engine.

    Example:
    -------
        >>> process_engine("foobar.db")
        >>> process_engine("postgresql+psycopg2://<user>:<password>@<host>/<db>"

    """

    group = parser.add_argument_group("database", "Database related options")
    group.add_argument("--db", dest="engine", help="sqlalchemy uri or filename")
    group.add_argument(
        "--db-create", action="store_true", help="create the db structures"
    )

    def callback(args: argparse.Namespace):
        from sqlalchemy import create_engine
        from sqlalchemy.engine.url import make_url
        from sqlalchemy.exc import ArgumentError

        if not args.engine or not args.engine.strip():
            return None

        engine = None
        with contextlib.suppress(ArgumentError):
            url = make_url(args.engine)
            engine = create_engine(url)
        args.engine = engine or create_engine(f"sqlite:///{args.engine}", echo=False)

    parser.callbacks.append(callback)


def add_arguments_new_miners_ips(parser: LuxosParserBase):
    group = parser.add_argument_group("Miners", "miners list or range")
    group.add_argument(
        "--range",
        action="append",
        dest="addresses",
        help="IPs range or @file",
        type=type_range,
    )
    group.add_argument(
        "--port", dest="port", help="miners' default port", type=int, default=4028
    )

    def callback(args: argparse.Namespace):
        addresses = []
        if args.addresses:
            addresses.extend(
                [
                    (host, port or args.port)
                    for group in args.addresses
                    for host, port in group
                ]
            )
        args.addresses = addresses

    parser.callbacks.append(callback)



from __future__ import annotations

import io


def indent(txt: str, pre: str = " " * 2) -> str:
    """simple text indentation"""

    from textwrap import dedent

    txt = dedent(txt)
    if txt.endswith("\n"):
        last_eol = "\n"
        txt = txt[:-1]
    else:
        last_eol = ""

    result = pre + txt.replace("\n", "\n" + pre) + last_eol
    return result if result.strip() else result.strip()


def md(txt: str, md: bool | None = None, **kwargs) -> str:
    """convert text to markdown if rich library is present"""
    try:
        from rich.markdown import Markdown

        md = True if md is None else md
    except ModuleNotFoundError:
        md = False

    if md:
        from rich.console import Console

        buf = io.StringIO()
        console = Console(file=buf, **kwargs)
        console.print(Markdown(indent(txt, "")))
        return buf.getvalue()
    else:
        return indent(txt, "")



"""
A script to perform a health check on every LuxOS miner in the network.
"""

import time
import logging
import argparse
import os
import csv
import json
import threading
import asyncio
from datetime import datetime
from typing import Any

from tqdm.asyncio import tqdm as async_tqdm
import asyncpg
import shutil
import yaml
import pandas as pd

from luxos.api import logon_required

from luxos.scripts.luxos import (generate_ip_range,
                   add_session_id_parameter, parameters_to_string,
                   check_res_structure, get_str_field)

LOGGING_CONFIG = {
    'level': logging.INFO,
    'format': "%(asctime)s [%(levelname)s] %(message)s",
    'handlers':
    [logging.StreamHandler(),
     logging.FileHandler("LuxOS-CLI.log")],
}

COLUMN_NAMES = [
    'ip', 'model', 'mac_address', 'os', 'board', 'version', 'elapsed',
    'board_count', 'm_5_hashrate', 'atm_enabled', 'chips_healthy',
    'chips_unhealthy_count', 'board0_5m', 'board1_5m', 'board2_5m',
    'board0_profile', 'board1_profile', 'board2_profile', 'board0_temp',
    'board1_temp', 'board2_temp', 'pool_user', 'diff_accept_0',
    'diff_reject_0', 'diff_stale_0', 'diff_accept_1', 'diff_reject_1',
    'diff_stale_1', 'diff_accept_2', 'diff_reject_2', 'diff_stale_2'
]

BUFFER_SIZE = 100
MAX_RETRIES = 3

INTERMEDIATE_CSV = "report_in_progress.csv"

LUXOS_MINERS = []

log = logging.getLogger(__name__)


def parse_args():
    """Parse arguments from config file."""
    path = "config.yaml"
    with open(path, 'r') as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError:
            log.exception("Unable to load configs from %s!", path)
            raise

    args = argparse.Namespace()

    args.verbose = config['output']['verbose']
    args.csv_output = config['output']['csv_output']
    args.local_grafana_file = config['output']['local_grafana_file']
    args.report_output_type = config['output']['report_output_type']
    args.ipfile = config['ipfile']
    args.range_start = config['ip_settings']['range_start']
    args.range_end = config['ip_settings']['range_end']
    args.max_threads = config['threads']['max_threads']
    args.timeout = config['luxos']['timeout']
    args.port = config['luxos']['port']
    args.port = config['luxos']['port']
    args.sleep_between_executions = config['sleep_between_executions']
    args.db_host = config['database']['host']
    args.db_port = config['database']['port']
    args.db_database = config['database']['name']
    args.db_user = config['database']['user']
    args.db_password = config['database']['password']
    args.db_table_name = config['database']['table_name']
    args.executeconfigs = config['execution']['executeconfigs']
    args.executeconfig_tempctrlset = config['execution']['tempctrlset'][
        'execute']
    args.executeconfig_tempctrlset_param1 = config['execution']['tempctrlset'][
        'param1']
    args.executeconfig_tempctrlset_param2 = config['execution']['tempctrlset'][
        'param2']
    args.executeconfig_tempctrlset_param3 = config['execution']['tempctrlset'][
        'param3']
    args.executeconfig_atmset = config['execution']['atmset']['execute']
    args.executeconfig_atmset_param1 = config['execution']['atmset']['param1']
    args.executeconfig_atmset_param2 = config['execution']['atmset']['param2']
    args.executeconfig_atmset_param3 = config['execution']['atmset']['param3']
    args.executeconfig_atmset_param4 = config['execution']['atmset']['param4']
    args.executeconfig_fanset = config['execution']['fanset']['execute']
    args.executeconfig_fanset_param1 = config['execution']['fanset']['param1']
    args.executeconfig_fanset_param2 = config['execution']['fanset']['param2']
    args.executeconfig_immersionswitch = config['execution'][
        'immersionswitch']['execute']
    args.executeconfig_immersionswitch_param1 = config['execution'][
        'immersionswitch']['param1']

    return args


def generate_filename(base_name="output"):
    current_time = datetime.now()
    timestamp = current_time.strftime("%Y_%m_%d_%H%M%S")

    if '.' in base_name:
        name, extension = os.path.splitext(base_name)
    else:
        name = base_name
        extension = ".csv"

    return f"health_checks/{name}_{timestamp}{extension}"


def load_ip_list_from_csv(csv_file):
    if not os.path.exists(csv_file):
        logging.info(f"Error: {csv_file} file not found.")
        exit(1)
    df = pd.read_csv(csv_file)
    if 'hostname' not in df.columns:
        logging.info("Error: No 'hostname' column found in CSV.")
        exit(1)
    return df['hostname'].dropna().tolist()


def get_value_with_default(dictionary, key, default=0):
    """
    Retrieve a value from a dictionary for the given key.
    If the key is not present or the value is None, return the default value.
    """
    return dictionary.get(
        key, default) if dictionary.get(key) is not None else default


async def internal_send_cgminer_command(host: str, port: int, command: str,
                                        timeout_sec: int,
                                        verbose: bool) -> dict[str, Any]:
    writer = None
    try:
        reader, writer = await asyncio.wait_for(asyncio.open_connection(
            host, port),
                                                timeout=timeout_sec)
        writer.write(command.encode())
        await writer.drain()

        response = bytearray()
        while True:
            data = await asyncio.wait_for(reader.read(1), timeout=timeout_sec)
            if not data:
                break
            null_index = data.find(b'\x00')
            if null_index >= 0:
                response += data[:null_index]
                break
            response += data

        r = json.loads(response.decode())
        if verbose:
            logging.info(r)
        return r

    except asyncio.TimeoutError:
        logging.error(f"Timeout on {host}.")
        return {}
    except Exception as e:
        logging.error(f"Error during socket operation: {e}")
        return {}
    finally:
        if writer:
            writer.close()
            await writer.wait_closed()


async def send_cgminer_command(host: str, port: int, cmd: str, param: str,
                               timeout_sec: int, verbose: bool) -> dict[str, Any]:
    req = str(f"{{\"command\": \"{cmd}\", \"parameter\": \"{param}\"}}\n")
    if verbose:
        logging.info(
            f"Executing command: {cmd} with params: {param} to host: {host}")

    return await internal_send_cgminer_command(host, port, req, timeout_sec,
                                               verbose)


async def send_cgminer_simple_command(host: str, port: int, cmd: str,
                                      timeout: int, verbose: bool) -> dict[str, Any]:
    req = str(f"{{\"command\": \"{cmd}\"}}\n")
    if verbose:
        logging.info(f"Executing command: {cmd} to host: {host}")
    return await internal_send_cgminer_command(host, port, req, timeout,
                                               verbose)


async def logon(host: str, port: int, timeout: int, verbose: bool) -> str:
    res = await send_cgminer_simple_command(host, port, "logon", timeout,
                                            verbose)
    check_res_structure(res, "SESSION", 1, 1)
    session = res["SESSION"][0]
    s = get_str_field(session, "SessionID")
    if s == "":
        raise ValueError("error: invalid session id")
    return s


async def execute_command(host: str, port: int, timeout_sec: int, cmd: str,
                          parameters: list, verbose: bool):
    try:
        logon_req = logon_required(cmd)
        if logon_req:
            sid = await logon(host, port, timeout_sec, verbose)
            parameters = add_session_id_parameter(sid, parameters)
            if verbose:
                logging.info(
                    f'Command requires a SessionID, logging in for host: {host}'
                )
                logging.info(f'SessionID obtained for {host}: {sid}')
        # TODO fix this typing issue
        elif not logon_required and verbose:  # type: ignore
            logging.info(f"Logon not required for executing {cmd}")
        param_string = parameters_to_string(parameters)
        if verbose:
            logging.info(f"{cmd} on {host} with parameters: {param_string}")
        res = await send_cgminer_command(host, port, cmd, param_string,
                                         timeout_sec, verbose)

        if verbose:
            logging.info(res)
        if logon_req:
            await send_cgminer_command(host, port, "logoff", sid, timeout_sec,
                                       verbose)

        return res
    except Exception as e:
        logging.error(f"Error executing {cmd} on {host}: {e}")


def parse_devs(json):
    devs_data = {
        'alive_board_count': 0,
        '5m_hr': 0,
        'elapsed': json.get('DEVS', [{}])[0].get('Device Elapsed', 0) / 60,
        'boards': []
    }

    all_boards = json.get('DEVS', [])
    for i in range(3):
        board = all_boards[i] if i < len(all_boards) else {}

        board_data = {
            '5m_hr': get_value_with_default(board, 'MHS 5m', 0) / 1e6,
            'temperature': board.get('Temperature', 0),
            'profile': board.get('Profile', 'Unknown')
        }

        devs_data['boards'].append(board_data)

        if board.get('Status') == 'Alive':
            devs_data['alive_board_count'] += 1
            devs_data['5m_hr'] += board_data['5m_hr']

    return devs_data


def parse_board(fgpabuildhex, fpgabuildidstr):
    if fgpabuildhex.lower().startswith('0xbbb'):
        return "BBB"
    elif fgpabuildhex == '0x000A113D':
        return "AML"
    elif ''.join(e for e in fpgabuildidstr
                 if e.isalnum())[2:-2] == fgpabuildhex[2:]:
        return "Xilinx"

    return "Unknown"


def parse_healthchip(healthchipget):
    try:
        unhealthy_count = 0
        for chip in healthchipget.get("CHIPS", []):
            if chip.get("Healthy") == "N":
                unhealthy_count += 1

        healthy = unhealthy_count == 0

        return {
            'chips_healthy': healthy,
            'chips_unhealthy_count': unhealthy_count
        }

    except Exception as e:
        logging.error(f"Error processing healthchipget data: {e}")
        return {}


def parse_config(json):
    json = json.get('CONFIG', [{}])[0]

    fgpabuildhex = json.get('FPGABuildIdHex', '')
    fpgabuildidstr = json.get('FPGABuildIdStr', '')
    if fgpabuildhex:
        board = parse_board(fgpabuildhex, fpgabuildidstr)

    config_data = {
        'is_atm_enabled': json.get('IsAtmEnabled', False),
        'model': json.get('Model', 'Unknown'),
        'mac_addr': json.get('MACAddr', 'Unknown'),
        'os': json.get('OS', 'Unknown'),
        'board': board
    }
    return config_data


def parse_version(json):
    try:
        luxminer_version = json.get('VERSION',
                                    [{}])[0].get('LUXminer', 'Unknown')
    except Exception:
        luxminer_version = 'Unknown'
    return {'version': luxminer_version}


def parse_stats(json):
    elapsed = json.get('STATS', [{}])[0].get('Elapsed', 0) / 60
    stats_data = {'elapsed': elapsed}
    return stats_data


def parse_pools(json):
    pool_data = {}
    pools = json.get('POOLS', [{}])

    try:
        pool_data['pool_user'] = pools[0].get('User', '')
    except Exception:
        pool_data['pool_user'] = ''

    for pool_id in range(3):
        try:
            pool = pools[pool_id]
            pool_data[f'diff_accept_{pool_id}'] = pool.get(
                'Difficulty Accepted', 'Unknown')
            pool_data[f'diff_reject_{pool_id}'] = pool.get(
                'Difficulty Rejected', 'Unknown')
            pool_data[f'diff_stale_{pool_id}'] = pool.get(
                'Difficulty Stale', 'Unknown')
        except Exception:
            pool_data[f'diff_accept_{pool_id}'] = 'Pool_ID out of range.'
            pool_data[f'diff_reject_{pool_id}'] = 'Pool_ID out of range.'
            pool_data[f'diff_stale_{pool_id}'] = 'Pool_ID out of range.'

    return pool_data


def parse_bitmain_stats(data):
    stats = data.get('STATS', [{}])

    while len(stats) < 2:
        stats.append({})
    j0 = stats[0]
    j1 = stats[1]

    stats_data = {
        'model':
        j0.get('Type', 'Unknown'),
        'MACAddress':
        'Unknown',
        'OS':
        'bitmain',
        'Elapsed':
        j1.get('Elapsed', 0) / 60,
        'Board Count':
        j1.get('miner_count', 0),
        '5M Hashrate': (float(j1.get('chain_rate1', 0) or 0) +
                        float(j1.get('chain_rate2', 0) or 0) +
                        float(j1.get('chain_rate3', 0) or 0)) / 1e3,
        'ATM Enabled':
        'Unknown',
        'Board05M':
        float(j1.get('chain_rate1', 0) or 0),
        'Board15M':
        float(j1.get('chain_rate2', 0) or 0),
        'Board25M':
        float(j1.get('chain_rate3', 0) or 0),
        'Board0 Profile':
        j1.get('freq1', 0),
        'Board1 Profile':
        j1.get('freq2', 0),
        'Board2 Profile':
        j1.get('freq3', 0),
        'Board0 Temp':
        j1.get('temp2_1', 0),
        'Board1 Temp':
        j1.get('temp2_2', 0),
        'Board2 Temp':
        j1.get('temp2_3', 0),
    }

    return stats_data


def generate_row(ip, devs, config, pools, version, healthchipget):
    row_data = [
        ip,
        config.get('model', ''),
        config.get('mac_addr', ''),
        config.get('os', ''),
        config.get('board', ''),
        version.get('version', ''),
        devs.get('elapsed', ''),
        devs.get('alive_board_count', ''),
        devs.get('5m_hr', ''),
        config.get('is_atm_enabled', '')
    ]
    healthchipget = [
        healthchipget.get('chips_healthy', ''),
        healthchipget.get('chips_unhealthy_count', '')
    ]
    pool_data = [
        pools.get('pool_user', ''),
        pools.get('diff_accept_0', ''),
        pools.get('diff_reject_0', ''),
        pools.get('diff_stale_0', ''),
        pools.get('diff_accept_1', ''),
        pools.get('diff_reject_1', ''),
        pools.get('diff_stale_1', ''),
        pools.get('diff_accept_2', ''),
        pools.get('diff_reject_2', ''),
        pools.get('diff_stale_2', '')
    ]
    boards = devs.get('boards', [])
    if boards and not isinstance(boards, list):
        boards = [boards]

    board_5m_hrs = [board.get('5m_hr', '') for board in boards]
    board_profiles = [board.get('profile', '') for board in boards]
    board_temps = [board.get('temperature', '') for board in boards]

    return row_data + healthchipget + board_5m_hrs + board_profiles + board_temps + pool_data


def generate_empty_row(ip):
    return [ip] + ['' for _ in COLUMN_NAMES[1:]]


def generate_bitmain_row(ip, stats, pools):
    healthchip = {}
    version = {}

    stats_data = [
        ip,
        stats.get('model', ''),
        stats.get('MACAddress', ''),
        stats.get('OS', ''),
        stats.get('Board', ''),
        version.get('Version', ''),
        stats.get('Elapsed', ''),
        stats.get('Board Count', ''),
        stats.get('5M Hashrate', ''),
        stats.get('ATM Enabled', ''),
        healthchip.get('chips_healthy', ''),
        healthchip.get('chips_unhealthy_count', ''),
        stats.get('Board05M', ''),
        stats.get('Board15M', ''),
        stats.get('Board25M', ''),
        stats.get('Board0 Profile', ''),
        stats.get('Board1 Profile', ''),
        stats.get('Board2 Profile', ''),
        stats.get('Board0 Temp', ''),
        stats.get('Board1 Temp', ''),
        stats.get('Board2 Temp', '')
    ]

    pool_data = [
        pools.get('pool_user', ''),
        pools.get('diff_accept_0', ''),
        pools.get('diff_reject_0', ''),
        pools.get('diff_stale_0', ''),
        pools.get('diff_accept_1', ''),
        pools.get('diff_reject_1', ''),
        pools.get('diff_stale_1', ''),
        pools.get('diff_accept_2', ''),
        pools.get('diff_reject_2', ''),
        pools.get('diff_stale_2', '')
    ]

    return stats_data + pool_data


async def create_connection_pool(db_config):
    return await asyncpg.create_pool(**db_config)


async def insert_to_db(pool, rows_to_insert, table_name):
    async with pool.acquire() as conn:
        async with conn.transaction():
            for row in rows_to_insert:
                converted_row = [str(item) for item in row]

                await conn.execute(
                    f'''
                    INSERT INTO {table_name}(
                        ip, model, mac_address, os, board, version, elapsed, board_count, 
                        m_5_hashrate, atm_enabled, chips_healthy, chips_unhealthy_count, board0_5m, board1_5m, board2_5m, 
                        board0_profile, board1_profile, board2_profile, board0_temp, 
                        board1_temp, board2_temp, pool_user, diff_accept_0, diff_reject_0, 
                        diff_stale_0, diff_accept_1, diff_reject_1, diff_stale_1, 
                        diff_accept_2, diff_reject_2, diff_stale_2
                    ) VALUES(
                        $1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, 
                        $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28, $29, $30, $31
                    )
                    ''', *converted_row)


async def append_to_buffer_and_check(buffer, row, lock, pool, table_name):
    rows_to_insert = []
    async with lock:
        buffer.append(row)
        if len(buffer) >= BUFFER_SIZE:
            rows_to_insert = list(buffer)
            buffer.clear()

    if rows_to_insert:
        await insert_to_db(pool, rows_to_insert, table_name)


async def handle_csv_row(row, lock, buffer, extra_arg=None, table_name=None):
    async with lock:
        buffer.append(row)

        if len(buffer) >= BUFFER_SIZE:
            with open(INTERMEDIATE_CSV, 'a', newline='') as csvfile:
                writer = csv.writer(csvfile)
                writer.writerows(buffer)
            del buffer[:]


async def handle_db_row(row, lock, buffer, pool, table_name):
    await append_to_buffer_and_check(buffer, row, lock, pool, table_name)


async def execute_with_retries(ip, port, timeout, command_name, verbose):
    for retry in range(1, MAX_RETRIES + 1):
        result = await execute_command(ip, port, timeout * retry, command_name,
                                       '', verbose)
        if result and isinstance(result, dict):
            return result
        logging.warning(
            f"Attempt {retry} failed for {ip}, command {command_name}. Retrying..."
        )
    return None


async def handle_devs_and_config(ip, res_devs, args, lock, buffer,
                                 handle_row_func, extra_arg):
    devs = parse_devs(res_devs)
    res_config = await execute_with_retries(ip, args.port, args.timeout,
                                            'config', args.verbose)
    if not res_config:
        logging.warning(
            f"Failed to get config for {ip} after {MAX_RETRIES} attempts. Skipping."
        )
        return
    config = parse_config(res_config)

    res_pools = await execute_with_retries(ip, args.port, args.timeout,
                                           'pools', args.verbose)

    if not res_pools:
        logging.warning(
            f"Failed to get pools for {ip} after {MAX_RETRIES} attempts. Skipping."
        )
        return

    res_version = await execute_with_retries(ip, args.port, args.timeout,
                                             'version', args.verbose)

    if not res_version:
        logging.warning(
            f"Failed to get version for {ip} after {MAX_RETRIES} attempts. Skipping."
        )
        return

    res_healthchipget = await execute_with_retries(ip, args.port, args.timeout,
                                                   'healthchipget',
                                                   args.verbose)

    if not res_healthchipget:
        logging.warning(
            f"Failed to get healthchipget for {ip} after {MAX_RETRIES} attempts. Skipping."
        )
        return

    healthchipget = parse_healthchip(res_healthchipget)
    version = parse_version(res_version)
    pools = parse_pools(res_pools)
    new_row = generate_row(ip, devs, config, pools, version, healthchipget)
    await handle_row_func(new_row, lock, buffer, extra_arg, args.db_table_name)


async def handle_status_and_stats(ip, args, lock, buffer, handle_row_func,
                                  extra_arg):
    res_bitmain = await execute_with_retries(ip, args.port, args.timeout,
                                             'stats', args.verbose)
    if not res_bitmain:
        logging.warning(
            f"Failed to get stats for {ip} after {MAX_RETRIES} attempts. Skipping."
        )
        return

    res_pools = await execute_with_retries(ip, args.port, args.timeout,
                                           'pools', args.verbose)
    if not res_pools:
        logging.warning(
            f"Failed to get pools for {ip} after {MAX_RETRIES} attempts. Skipping."
        )
        return

    pools = parse_pools(res_pools)

    stats = parse_bitmain_stats(res_bitmain)
    new_row = generate_bitmain_row(ip, stats, pools)
    await handle_row_func(new_row, lock, buffer, extra_arg, args.db_table_name)


async def handle_unknown_response(ip, args, res_devs, lock, buffer,
                                  handle_row_func, extra_arg):
    logging.error(f'Unknown API response from {ip}: {res_devs}')
    new_row = generate_empty_row(ip)
    await handle_row_func(new_row, lock, buffer, extra_arg, args.db_table_name)


async def healthcheck(ip,
                      handle_row_func,
                      buffer,
                      lock,
                      args,
                      sem,
                      extra_arg=None):
    async with sem:
        res_devs = await execute_with_retries(ip, args.port, args.timeout,
                                              'devs', args.verbose)
        if not res_devs:
            logging.warning(
                f"Failed to get a response for {ip} after {MAX_RETRIES} attempts. Skipping."
            )
            new_row = generate_empty_row(ip)
            await handle_row_func(new_row, lock, buffer, extra_arg,
                                  args.db_table_name)
            return

        keys = list(res_devs.keys())

        if keys == ['DEVS', 'STATUS', 'id']:
            await handle_devs_and_config(ip, res_devs, args, lock, buffer,
                                         handle_row_func, extra_arg)

            if args.executeconfigs == 'True':
                async with lock:
                    LUXOS_MINERS.append(ip)
        elif keys == ['STATUS', 'DEVS', 'id']:
            await handle_status_and_stats(ip, args, lock, buffer,
                                          handle_row_func, extra_arg)

        else:
            await handle_unknown_response(ip, args, res_devs, lock, buffer,
                                          handle_row_func, extra_arg)


async def execute_configs(ip, args):
    if args.executeconfig_atmset == 'True':
        try:
            await execute_command(ip, args.port, args.timeout, 'atmset', [
                args.executeconfig_atmset_param1,
                args.executeconfig_atmset_param2,
                args.executeconfig_atmset_param3,
                args.executeconfig_atmset_param4
            ], args.verbose)
            logging.info(f"Executed atmset config on {ip}")
        except Exception as e:
            logging.error(f"Failed to execute atmset config on {ip}: {e}")

    if args.executeconfig_fanset == 'True':
        try:
            await execute_command(ip, args.port, args.timeout, 'fanset', [
                str(args.executeconfig_fanset_param1),
                str(args.executeconfig_fanset_param2)
            ], args.verbose)
            logging.info(f"Executed fanset config on {ip}")
        except Exception as e:
            logging.error(f"Failed to execute fanset config on {ip}: {e}")

    if args.executeconfig_tempctrlset == 'True':
        try:
            await execute_command(ip, args.port, args.timeout, 'tempctrlset', [
                str(args.executeconfig_tempctrlset_param1),
                str(args.executeconfig_tempctrlset_param2),
                str(args.executeconfig_tempctrlset_param3)
            ], args.verbose)
            logging.info(f"Executed tempctrlset config on {ip}")
        except Exception as e:
            logging.error(f"Failed to execute tempctrlset config on {ip}: {e}")

    if args.executeconfig_immersionswitch == 'True':
        try:
            await execute_command(
                ip, args.port, args.timeout, 'immersionswitch',
                [str(args.executeconfig_immersionswitch_param1)], args.verbose)
            logging.info(f"Executed immersionswitch config on {ip}")
        except Exception as e:
            logging.error(
                f"Failed to execute immersionswitch config on {ip}: {e}")


async def main_loop(ip_list, args, lock, sem, buffer, db_config):
    health_checks_folder = os.path.join(os.getcwd(), "health_checks")
    if not os.path.exists(health_checks_folder):
        os.makedirs(health_checks_folder)

    while True:
        LUXOS_MINERS.clear()
        buffer.clear()
        logging.info("Starting HealthCheck...")
        start_time = time.time()

        start_time_healthcheck = time.time()
        if args.report_output_type == 'csv':
            try:
                with open(INTERMEDIATE_CSV, 'w', newline='') as csvfile:
                    writer = csv.writer(csvfile)
                    writer.writerow(COLUMN_NAMES)
                tasks = [
                    healthcheck(ip, handle_csv_row, buffer, lock, args, sem)
                    for ip in ip_list
                ]
                for task in async_tqdm.as_completed(tasks,
                                                    desc="Processing IPs",
                                                    total=len(tasks)):
                    await task
                if buffer:
                    with open(INTERMEDIATE_CSV, 'a', newline='') as csvfile:
                        writer = csv.writer(csvfile)
                        writer.writerows(buffer)
                final_output_name = generate_filename(args.csv_output)
                shutil.copy(INTERMEDIATE_CSV, args.local_grafana_file)
                os.rename(INTERMEDIATE_CSV, final_output_name)

            finally:
                if os.path.exists(INTERMEDIATE_CSV):
                    os.remove(INTERMEDIATE_CSV)
        else:
            pool = await create_connection_pool(db_config)
            try:
                tasks = [
                    healthcheck(ip, handle_db_row, buffer, lock, args, sem,
                                pool) for ip in ip_list
                ]
                for task in async_tqdm.as_completed(tasks,
                                                    desc="Processing IPs",
                                                    total=len(tasks)):
                    await task
                if buffer:
                    await insert_to_db(pool, buffer, args.db_table_name)
            finally:
                await pool.close()

        end_time_healthcheck = time.time()
        execution_time_healthcheck = end_time_healthcheck - start_time_healthcheck
        logging.info("Finished performing HealthCheck on all hosts.")
        logging.info(
            f"Execution time for Health Check: {execution_time_healthcheck:.2f} seconds."
        )

        if args.executeconfigs == 'True':
            logging.info("Starting config settings...")
            start_time_config = time.time()
            tasks = [execute_configs(ip, args) for ip in LUXOS_MINERS]
            for task in async_tqdm.as_completed(tasks,
                                                desc="Executing configs",
                                                total=len(tasks)):
                await task
            end_time_config = time.time()
            execution_time_config = end_time_config - start_time_config
            logging.info(
                f"Execution time for configs: {execution_time_config:.2f} seconds."
            )

        end_time = time.time()
        execution_time = end_time - start_time

        logging.info(f"Total execution time: {execution_time:.2f} seconds.")

        await asyncio.sleep(args.sleep_between_executions)


def print_message_on_timer():
    while True:
        print(" ########## PRESS ENTER TO STOP ##########")
        time.sleep(20)


def stop_on_input(loop):
    message_thread = threading.Thread(target=print_message_on_timer,
                                      daemon=True)
    message_thread.start()
    input()
    print("Cancelling...")
    for task in asyncio.all_tasks(loop):
        task.cancel()
    message_thread.join()


async def run():
    try:
        args = parse_args()
        logging.basicConfig(**LOGGING_CONFIG)

        db_config = {
            'user': args.db_user,
            'password': args.db_password,
            'database': args.db_database,
            'host': args.db_host,
            'port': args.db_port
        }

        lock = asyncio.Lock()
        sem = asyncio.Semaphore(args.max_threads)
        buffer = []

        if args.range_start and args.range_end:
            ip_list = generate_ip_range(args.range_start, args.range_end)
        elif args.ipfile:
            ip_list = load_ip_list_from_csv(args.ipfile)
        else:
            logging.info("No IP address or IP list found.")
            exit(1)

        loop = asyncio.get_event_loop()
        threading.Thread(target=stop_on_input, args=(loop, ),
                         daemon=True).start()

        await main_loop(ip_list, args, lock, sem, buffer, db_config)
    except asyncio.CancelledError:
        logging.info("Interrupted by user. Exiting...")
        return


def main():
    asyncio.run(run())


if __name__ == "__main__":
    main()


"""ipaddress manipulation"""

from __future__ import annotations

import ipaddress
import re
from pathlib import Path
from typing import Generator

from .exceptions import AddressParsingError, LuxosBaseException


class DataParsingError(LuxosBaseException):
    pass


def splitip(txt: str) -> tuple[str, int | None]:
    expr = re.compile(r"(?P<ip>\d{1,3}([.]\d{1,3}){3})(:(?P<port>\d+))?")
    if not (match := expr.search(txt)):
        raise RuntimeError(f"invalid ip:port address {txt}")
    return match["ip"], int(match["port"]) if match["port"] is not None else None


def parse_expr(txt: str) -> None | tuple[str, str | None, int | None]:
    """parse text into a (start, end, port) tuple.

    Parses a text in these forms:
        <ip>
        <ip>:port
        <startip>:<endip>
        <startip>:<port>:<endip>
    Eg.
        >>> parse_expr("127.0.0.1")
        ("127.0.0.1", None, None)
        >>> ips.parse_expr("127.0.0.1:1234:127.0.0.3")
        ("127.0.0.1", "127.0.0.3", 1234)
    """
    tokens = {
        "ip": re.compile(r"(?P<ip>\d{1,3}([.]\d{1,3}){3})"),
        "sep": re.compile(":"),
        "div": re.compile("-"),
        "port": re.compile(r"(?P<port>\d+)"),
        "address": re.compile(r"(?P<address>[^:]+)"),
    }

    txt2 = txt.replace(" ", "")

    items = []
    while txt2.strip():
        for k, e in tokens.items():
            if match := e.match(txt2):
                i, j = match.span()
                items.append((k, txt2[i:j]))
                txt2 = txt2[j:]
                break
        else:
            raise AddressParsingError(f"cannot parse text '{txt}'")

    if len(items) == 0:
        raise AddressParsingError(f"cannot parse '{txt}'")

    def matcher(syntax):
        def match(left, right):
            if len(left) != len(right):
                return False
            for a, b in zip(left, right):
                if isinstance(b, str):
                    if a != b:
                        return False
                elif a not in b:
                    return False
            return True

        start = end = port = None
        if match(syntax, [{"ip", "address"}]):
            start = items[0][1]
        elif match(syntax, [{"ip", "address"}, "sep", "port"]):
            start = items[0][1]
            port = int(items[2][1])
        elif match(syntax, ["ip", {"sep", "div"}, "ip"]):
            start = items[0][1]
            end = items[2][1]
        elif match(syntax, ["ip", "sep", "port", {"div", "sep"}, "ip"]):
            start = items[0][1]
            end = items[4][1]
            port = int(items[2][1])
        elif match(syntax, ["ip", {"div", "sep"}, "ip", "sep", "port"]):
            start = items[0][1]
            end = items[2][1]
            port = int(items[4][1])
        elif match(syntax, ["ip", "sep", "port", {"div", "sep"}, "ip", "sep", "port"]):
            start = items[0][1]
            end = items[4][1]
            port = int(items[2][1])
            port1 = int(items[6][1])
            if port != port1:
                raise AddressParsingError(f"ports mismatch {port} != {port1}")
        else:
            raise AddressParsingError(f"cannot parse '{txt}': {syntax=}")
        return start, end, port

    syntax = [item[0] for item in items]
    return matcher(syntax)


def iter_ip_ranges(
    txt: str, port: int | None = None, gsep: str = ",", strict: bool = True
) -> Generator[tuple[str, int | None], None, None]:
    """iterate over ip ranges.

    The txt string cav have one of these formats:

    1. a single ip and (optional) port: ``127.0.0.1`` or ``127.0.0.1:8080``
    2. an (inclusive) range using two ips separated by a
       ``-`` (minus) sign: ``127.0.0.1-127.0.0.3``
    3. a combination of the above separated by a ``,`` (comma) sign:
       ``127.0.0.1,192.168.0.1-192.168.0.10``

    Example::

        for ip in iter_ip_ranges("127.0.0.1,127.0.0.3-127.0.0.15:9999"):
            print(ip)

        (127.0.0.1, None),
        (127.0.0.3, 9999),
        ...
        (127.0.0.15, 9999),
    """
    for segment in txt.replace(" ", "").split(gsep):
        try:
            if not (found := parse_expr(segment)):
                continue
        except AddressParsingError:
            if strict:
                raise
            continue
        start, end, theport = found
        if start is None and end is None:
            raise RuntimeError(f"cannot parse '{segment}'")
        if end is None:
            yield (start, theport or port)
            continue

        cur = ipaddress.IPv4Address(start)
        last = ipaddress.IPv4Address(end)
        while cur <= last:
            yield (str(cur), theport or port)
            cur += 1


def ip_ranges(
    txt: str, gsep: str = ":", strict: bool = True
) -> list[tuple[str, int | None]]:
    """return a list of ips given a text expression.

    Eg.
        >>> for ip in ip_ranges("127.0.0.1"):
        ...     print(ip)
        127.0.0.1

        >>> for ip in ip_ranges("127.0.0.1-127.0.0.3"):
        ...     print(ip)
        127.0.0.1
        127.0.0.2
        127.0.0.3

    NOTE: use the `:` (gsep) to separate ips groups, and `-` (rsep) to define a range.
    """
    return list(iter_ip_ranges(txt, gsep=gsep, strict=strict))


def load_ips_from_csv(
    path: Path | str, port: int | None = 4028, strict: bool = False
) -> list[tuple[str, int | None]]:
    """
    Load ip addresses from a csv file.

    Arguments:
        path: a Path object to load data from (csv-like)
        port: a fallback port if not defined
        strict: abort with AddressParsingError if there's a malformed entry

    Raises:
        AddressParsingError: if strict is set to True and there's a
            malformed entry in path.

    Notes:
        The **strict** argument if set to False will ignore malformed
        lines in csv. If set to True it will raise AddressParsingError on
        invalid entries.

    Example:
        **foobar.csv** file::

            # comment (or empty lines) will be ignored
            127.0.0.1 # a single address
            127.0.0.2-127.0.0.10 # a range of addresses

            # you can specify a port
            127.0.0.11:9999
            127.0.0.12-127.0.0.20:8888

        You can read into a list of (host, port) tuples as::

           for ip in load_ips_from_csv("foobar.csv"):
               print(ip)

           (127.0.0.1, 4028)
           (127.0.0.2, 4028)
           (127.0.0.3, 4028)
           ...
           (127.0.0.10, 4028)

    """
    result = []
    for line in Path(path).read_text().split("\n"):
        line = line.partition("#")[0]
        if not line.strip():
            continue
        # for excel, an exception
        if line.strip().lower() == "hostname":
            continue
        for host, port2 in iter_ip_ranges(line, strict=strict):
            result.append((host, port2 or port))
    return result


def load_ips_from_yaml(
    path: Path | str, port: int | None = 4028, strict: bool = False
) -> list[tuple[str, int | None]]:
    """
    Load ip addresses from a yaml file.

    Arguments:
        path: a Path object to load data from (csv-like)
        port: a fallback port if not defined
        strict: abort with AddressParsingError if there's a malformed entry

    Raises:
        AddressParsingError: if strict is set to True and
            there's a malformed entry in path.

    Notes:
        The **strict** argument if set to False will ignore malformed
        lines in csv. If set to True it will raise AddressParsingError on
        invalid entries.

    Example:
        **foobar.yaml** file::

            miners:
                luxos_port: 9999  # default fallback port
                addresses:
                    - 127.0.0.1 # a single address
                    - 127.0.0.2-127.0.0.10 # a range of addresses

                    # you can specify a port
                    - 127.0.0.11:9999
                    - 127.0.0.12-127.0.0.20:8888
    """
    from yaml import safe_load

    txt = Path(path).read_text()
    try:
        data = safe_load(txt)
    except Exception as exc:
        raise DataParsingError(f"cannot parse yaml file {path}") from exc

    if "miners" in data and "addresses" in data["miners"]:
        miners = data["miners"]
        default_port = miners.get("luxos_port", None)
        result = []
        for address in data["miners"]["addresses"]:
            for host, thisport in iter_ip_ranges(address, strict=strict):
                result.append((host, thisport or default_port or port))
        return result

    raise DataParsingError(f"cannot find miners definitions in {path}")



from __future__ import annotations

import asyncio
import dataclasses as dc
import json
import traceback
from typing import Any

from .. import asyncops, misc, text


@dc.dataclass
class Result:
    host: str
    port: int
    tback: str = ""
    value: Any = None


async def wrapper(fn, host: str, port: int, *args, **kwargs) -> Result:
    try:
        return Result(host, port, value=await fn(host, port, *args, **kwargs))
    except asyncio.TimeoutError:
        return Result(host, port, tback="timeout error")
    except Exception:
        return Result(host, port, tback="".join(traceback.format_exc()))


async def run(
    ipaddresses: list[tuple[str, int]],
    cmd: str,
    params: list[str] | dict[str, str] | None,
    delay: float | None,
    details: str,
    batchsize: int = 0,
) -> None:
    result = {}

    for grupid, addresses in enumerate(misc.batched(ipaddresses, n=batchsize)):
        tasks = []
        for host, port in addresses:
            tasks.append(wrapper(asyncops.rexec, host, port, cmd, params))
        result[grupid] = await asyncio.gather(*tasks)

        # runs only on batchsize, wait delay then proceed onto the next batch
        if delay:
            await asyncio.sleep(delay)

    alltasks = [task for group in result.values() for task in group]
    successes = [task for task in alltasks if not task.tback]
    failures = [task for task in alltasks if task.tback]

    # print a nice report
    if details == "json":
        output = {}
        for task in successes:
            output[f"{task.host}:{task.port}"] = {"status": True, "result": task.value}
        for task in failures:
            output[f"{task.host}:{task.port}"] = {"status": False, "result": task.tback}
        print(json.dumps(output, indent=2, sort_keys=True))
    else:
        if details:
            print(f"task executed sucessfully: {len(successes)}")
            for task in successes:
                print(f"  > {task.host}:{task.port}")
                print(
                    text.indent(
                        json.dumps(task.value, indent=2, sort_keys=True), pre="  | "
                    )
                )
        else:
            print(
                f"task executed sucessfully (use -a|--all for details): "
                f"{len(successes)}"
            )

        if details:
            print(f"task executed with failures: {len(failures)}")
            for task in failures:
                print(f"  {task.host}:{task.port}:")
                print(f"  {text.indent(task.tback, '| ')}")
        else:
            print(
                f"task executed with failures (use -a|--all for details): "
                f"{len(failures)}"
            )


def main(*args, **kwargs) -> None:
    asyncio.run(run(*args, **kwargs))



from __future__ import annotations

import contextlib
import functools
import json
import logging
import socket
import time
from typing import Any

from luxos.api import logon_required

from . import exceptions
from .asyncops import (
    RETRIES,
    RETRIES_DELAY,
    TIMEOUT,
    parameters_to_list,
    validate,  # noqa: F401
    validate_message,
)
from .exceptions import MinerCommandSessionAlreadyActive, MinerConnectionError

log = logging.getLogger(__name__)


def wrapped(function):
    """wraps a function acting on a host and re-raise with internal exceptions

    This re-raise exceptions so they all derive from MinerConnectionError, eg:
    @wrapped
    def somcode(host: str, port: int, ...):
        ...

    try:
        await somecode()
    except MinerConnectionError as e:  <- this will catch all exceptions!
        e.address
        raise MyNewExecption() from e  <- this will re-raise
    """

    @functools.wraps(function)
    def _function(host: str, port: int, *args, **kwargs):
        try:
            return function(host, port, *args, **kwargs)
        except TimeoutError as e:
            # we augment underlying TimeOuts
            raise exceptions.MinerCommandTimeoutError(host, port) from e
        except exceptions.MinerConnectionError:
            raise
        except Exception as e:
            # we augment any other exception with (host, port) info
            log.exception("internal error")
            raise exceptions.MinerConnectionError(host, port, "internal error") from e

    return _function


def retryfn(
    timeout: float, retry: int, retry_delay: float, fn, host, port, *args, **kwargs
):
    """will retry fn calls up to timeout of up to number retry"""
    last_exception = None
    count = retry
    t0 = time.monotonic()
    while True:
        if retry > 0 and count <= 0:
            break
        if timeout and (time.monotonic() - t0) >= timeout:
            break

        try:
            return fn(*args, **kwargs)
        except Exception as e:
            last_exception = e
        log.debug("failed to retrieve result for '%s'", fn.__name__)

        if retry > 0:
            count -= 1
        if retry and retry_delay:
            time.sleep(retry_delay)

    if last_exception is not None:
        raise exceptions.MinerCommandTimeoutError(host, port) from last_exception


def _roundtrip(
    host: str, port: int, cmd: bytes | str, timeout: float | None = None
) -> str:
    """simple asyncio socket based send/receive function

    Example:
        print(_roundtrip(host, port, "version"))
        -> (str) "{'STATUS': [{'Code': 22, 'Description'...."
    """
    timeout = TIMEOUT if timeout is None else timeout
    # Create a socket connection to the server
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        # set timeout
        if timeout:
            sock.settimeout(timeout)

        # Connect to the server
        sock.connect((host, port))
        log.debug("connecting to %s:%i", host, port)
        # Send the command to the server
        sock.sendall(cmd.encode() if isinstance(cmd, str) else cmd)

        # Receive the response from the server
        response = []
        # Read one byte at a time so we can wait for the null terminator.
        # this is to avoid waiting for the timeout as we don't know how long
        # the response will be and socket.recv() will block until reading
        # the specified number of bytes.
        while data := sock.recv(2**3):
            response.append(data)

        result = "".join(block.decode() for block in response)
        log.debug("received: %s", result)
        return result


def roundtrip(
    host: str,
    port: int,
    cmd: bytes | str | dict[str, Any],
    asjson: bool | None = True,
    timeout: float | None = None,
    retry: int | None = 0,
    retry_delay: float | None = None,
):
    """utility wrapper around _roundrip

    Example:
        print(await roundtrip(host, port, {"version"}))
        -> (json) {'STATUS': [{'Code': 22, 'Description': 'LUXminer 20 ...
        print(await roundtrip(host, port, "version"))
        -> (str) "{'STATUS': [{'Code': 22, 'Description': 'LUXminer 20 ..
    """
    timeout = TIMEOUT if timeout is None else timeout
    retry = RETRIES if retry is None else retry
    retry_delay = RETRIES_DELAY if retry_delay is None else retry_delay

    if not isinstance(cmd, (bytes, str)):
        cmd = json.dumps(cmd, indent=2, sort_keys=True)
        if asjson is None:
            asjson = True

    last_exception = None
    count = retry
    t0 = time.monotonic()
    while True:
        if retry > 0 and count <= 0:
            break
        if timeout and (time.monotonic() - t0) >= timeout:
            break

        try:
            res = _roundtrip(host, port, cmd, timeout)
            if asjson:
                return json.loads(res)
            else:
                return res
        except Exception as e:
            last_exception = e
        log.debug("failed to retrieve result for '%s'", cmd)

        if retry > 0:
            count -= 1
        if retry and retry_delay:
            time.sleep(retry_delay)

    if last_exception is not None:
        raise exceptions.MinerCommandTimeoutError(host, port) from last_exception


@wrapped
def logon(host: str, port: int, timeout: float | None = None) -> str:
    timeout = TIMEOUT if timeout is None else timeout
    res = roundtrip(host, port, {"command": "logon"}, timeout=timeout)

    # when we first logon, we'll receive a token (session_id)
    #   [STATUS][SessionID]
    # on subsequent logon, we receive a
    #   [STATUS][Msg] == "Another session is active" ([STATUS][Code] 402)
    if "SESSION" not in res and res.get("STATUS", [{}])[0].get("Code") == 402:
        raise exceptions.MinerCommandSessionAlreadyActive(
            host, port, "connection active", res
        )
    sessions = validate_message(host, port, res, "SESSION", 1, 1)

    session = sessions["SESSION"][0]

    if "SessionID" not in session:
        raise exceptions.MinerCommandSessionAlreadyActive(
            host, port, "no SessionID in data", res
        )
    return str(session["SessionID"])


@wrapped
def logoff(
    host: str, port: int, sid: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    return roundtrip(
        host, port, {"command": "logoff", "parameter": sid}, timeout=timeout
    )


def rexec(
    host: str,
    port: int,
    cmd: str,
    parameters: str | int | float | bool | list[Any] | dict[str, Any] | None = None,
    timeout: float | None = None,
    retry: int | None = None,
    retry_delay: float | None = None,
) -> dict[str, Any]:
    parameters = parameters_to_list(parameters)

    timeout = TIMEOUT if timeout is None else timeout
    retry = RETRIES if retry is None else retry
    retry_delay = RETRIES_DELAY if retry_delay is None else retry_delay

    # if cmd is logon/logoff we dealt with it differently
    if cmd in {"logon", "logoff"}:
        failure = None
        for i in range(retry or 1):
            try:
                if cmd == "logon":
                    return {"sid": logon(host, port, timeout)}
                else:
                    return logoff(host, port, parameters[0])
            except Exception as exc:
                failure = exc
            if retry and (i < retry) and retry_delay:
                time.sleep(retry_delay)
        if isinstance(failure, Exception):
            raise failure

    failure = None
    sid = ""
    for i in range(retry + 1):
        if not logon_required(cmd):
            log.debug("no logon required for command '%s' on %s:%i", cmd, host, port)
            break
        try:
            sid = logon(host, port, timeout)
            parameters = [sid, *parameters]
            log.debug("session id requested & obtained for %s:%i (%s)", host, port, sid)
            break
        except Exception as exc:
            failure = exc
        if retry and (i < retry) and retry_delay:
            time.sleep(retry_delay)

    if isinstance(failure, Exception):
        raise failure

    packet = {"command": cmd}
    if parameters:
        packet["parameter"] = ",".join(parameters)
    log.debug(
        "executing command '%s' on '%s:%i' with parameters: %s",
        cmd,
        host,
        port,
        packet.get("parameter", ""),
    )

    failure = None
    for i in range(retry + 1):
        try:
            ret = roundtrip(host, port, packet, timeout=timeout)
            log.debug("received from %s:%s: %s", host, port, str(ret))
            if sid:
                logoff(host, port, sid)
            return ret
        except Exception as exc:
            failure = exc
        if retry and (i < retry) and retry_delay:
            log.debug("failed attempt %i (out of %i)", i + 1, retry)
            time.sleep(retry_delay)

    if sid:
        logoff(host, port, sid)
    if isinstance(failure, Exception):
        raise failure
    return {}


# !!! LEGACY CODE BELOW !!!


# internal_send_cgminer_command sends a command to the
# cgminer API server and returns the response.
def internal_send_cgminer_command(
    host: str, port: int, command: str, timeout: float | None
) -> dict[str, Any]:
    timeout_sec = TIMEOUT if timeout is None else timeout
    # Create a socket connection to the server
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            # set timeout
            if timeout_sec is not None:
                sock.settimeout(timeout)

            # Connect to the server
            sock.connect((host, port))

            # Send the command to the server
            sock.sendall(command.encode())

            # Receive the response from the server
            response = bytearray()
            while True:
                # Read one byte at a time so we can wait for the null terminator.
                # this is to avoid waiting for the timeout as we don't know how long
                # the response will be and socket.recv() will block until reading
                # the specified number of bytes.
                try:
                    data = sock.recv(1)
                except socket.timeout:
                    # Timeout occurred, check if we have any data so far
                    if len(response) == 0:
                        raise ValueError("timeout waiting for data")
                    else:
                        break
                if not data:
                    break
                null_index = data.find(b"\x00")
                if null_index >= 0:
                    response += data[:null_index]
                    break
                response += data

            # Parse the response JSON
            r = json.loads(response.decode())
            log.debug(r)
            return r

        except socket.error as e:
            raise e


# send_cgminer_command sends a command to the cgminer API server and
# returns the response.
def send_cgminer_command(
    host: str, port: int, cmd: str, param: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    req = str(f'{{"command": "{cmd}", "parameter": "{param}"}}\n')
    log.debug(f"Executing command: {cmd} with params: {param} to host: {host}")
    return internal_send_cgminer_command(host, port, req, timeout)


# send_cgminer_simple_command sends a command with no params
# to the miner and returns the response.
def send_cgminer_simple_command(
    host: str, port: int, cmd: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    req = str(f'{{"command": "{cmd}"}}\n')
    log.debug(f"Executing command: {cmd} to host: {host}")
    return internal_send_cgminer_command(host, port, req, timeout)


# check_res_structure checks that the response has the expected structure.
def check_res_structure(
    res: dict[str, Any], structure: str, min: int, max: int
) -> dict[str, Any]:
    # Check the structure of the response.
    if structure not in res or "STATUS" not in res or "id" not in res:
        raise ValueError("error: invalid response structure")

    # Should we check min and max?
    if min >= 0 and max >= 0:
        # Check the number of structure elements.
        if not (min <= len(res[structure]) <= max):
            raise ValueError(
                f"error: unexpected number of {structure} in response; "
                f"min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    # Should we check only min?
    if min >= 0:
        # Check the minimum number of structure elements.
        if len(res[structure]) < min:
            raise ValueError(
                f"error: unexpected number of {structure} in response; "
                f"min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    # Should we check only max?
    if max >= 0:
        # Check the maximum number of structure elements.
        if len(res[structure]) < min:
            raise ValueError(
                f"error: unexpected number of {structure} in response; "
                f"min: {min}, max: {max}, actual: {len(res[structure])}"
            )

    return res


# get_str_field tries to get the field as a string and returns it.
def get_str_field(struct: dict[str, Any], name: str) -> str:
    try:
        s = str(struct[name])
    except Exception as e:
        raise ValueError(f"error: invalid {name} str field: {e}")

    return s


def __logon(host: str, port: int, timeout: float | None = None) -> str:
    # Send 'logon' command to cgminer and get the response
    timeout = TIMEOUT if timeout is None else timeout
    res = send_cgminer_simple_command(host, port, "logon", timeout)
    if res["STATUS"][0]["STATUS"] == "E":
        raise MinerCommandSessionAlreadyActive(host, port, res["STATUS"][0]["Msg"])
    # Check if the response has the expected structure
    session = validate_message(host, port, res, "SESSION")[0]
    return str(session["SessionID"])


def __logoff(
    host: str, port: int, sid: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    res = send_cgminer_command(host, port, "logoff", sid, timeout)
    return validate_message(host, port, res)


def add_session_id_parameter(session_id, parameters) -> list[Any]:
    # Add the session id to the parameters
    return [session_id, *parameters]


def parameters_to_string(parameters):
    # Convert the parameters to a string that LuxOS API accepts
    return ",".join(parameters)


def execute_command(
    host: str,
    port: int,
    timeout_sec: int | float | None,
    cmd: str,
    parameters: str | int | float | bool | list[Any] | dict[str, Any] | None = None,
    verbose: bool = False,
):
    return rexec(host, port, cmd, parameters, timeout_sec)
    timeout_sec = TIMEOUT if timeout_sec is None else timeout_sec
    # Check if logon is required for the command
    logon_req = logon_required(cmd)

    parameters = parameters_to_list(parameters)

    if logon_req:
        # Get a SessionID
        sid = __logon(host, port, timeout_sec)
        # Add the SessionID to the parameters list at the left.
        parameters = add_session_id_parameter(sid, parameters)

        log.debug("SessionID obtained for %s: %s", host, sid)

    # TODO verify this
    elif not logon_req:
        log.debug("Logon not required for executing %s", cmd)

    # convert the params to a string that LuxOS API accepts
    param_string = ",".join(parameters)

    log.debug("%s on %s with parameters: %s", cmd, host, param_string)

    # Execute the API command
    res = send_cgminer_command(host, port, cmd, param_string, timeout_sec)

    log.debug(res)

    # Log off to terminate the session
    if logon_req:
        __logoff(host, port, sid, timeout_sec)

    return res


@contextlib.contextmanager
def with_atm(host, port, enabled: bool, timeout: float | None = None):
    res = rexec(host, port, "atm", timeout=timeout)
    if not res:
        raise MinerConnectionError(host, port, "cannot check atm")
    current = validate_message(host, port, res, "ATM")["ATM"][0]["Enabled"]
    rexec(host, port, "atmset", {"enabled": enabled}, timeout=timeout)
    yield current
    rexec(host, port, "atmset", {"enabled": current}, timeout=timeout)



from __future__ import annotations

import asyncio
import contextlib
import functools
import json
import logging
from typing import Any

from . import api, exceptions

log = logging.getLogger(__name__)

#: default timeout (s) for operations
TIMEOUT = 3.0
#: default number (>1) of retries on a failed operation
RETRIES = 0
#: delay (s) between retries
RETRIES_DELAY = 1.0


def wrapped(function):
    """wraps a function acting on a host and re-raise with internal exceptions

    This re-raise exceptions so they all derive from MinerConnectionError, eg:
    @wrapped
    def somcode(host: str, port: int, ...):
        ...

    try:
        await somecode()
    except MinerConnectionError as e:  <- this will catch all exceptions!
        e.address
        raise MyNewExecption() from e  <- this will re-raise
    """

    @functools.wraps(function)
    async def _function(host: str, port: int, *args, **kwargs):
        try:
            return await function(host, port, *args, **kwargs)
        except asyncio.TimeoutError as e:
            # we augment underlying TimeOuts
            raise exceptions.MinerCommandTimeoutError(host, port) from e
        except exceptions.MinerConnectionError:
            raise
        except Exception as e:
            # we augment any other exception with (host, port) info
            log.exception("internal error")
            raise exceptions.MinerConnectionError(host, port, "internal error") from e

    return _function


async def _roundtrip(
    host: str, port: int, cmd: bytes | str, timeout: float | None
) -> str:
    """simple asyncio socket based send/receive function

    Example:
        print(await _roundtrip(host, port, "version"))
        -> (str) "{'STATUS': [{'Code': 22, 'Description'...."
    """
    reader, writer = await asyncio.wait_for(
        asyncio.open_connection(host, port), timeout
    )

    writer.write(cmd.encode() if isinstance(cmd, str) else cmd)
    await writer.drain()

    response = bytearray()
    while True:
        data = await asyncio.wait_for(reader.read(1), timeout=timeout)
        if not data:
            break
        null_index = data.find(b"\x00")
        if null_index >= 0:
            response += data[:null_index]
            break
        response += data

    return response.decode()


# TODO add annotations
async def roundtrip(
    host: str,
    port: int,
    cmd: bytes | str | dict[str, Any],
    asjson: bool | None = True,
    timeout: float | None = None,
    retry: int | None = 0,
    retry_delay: float | None = None,
):
    """utility wrapper around _roundrip

    Example:
        print(await roundtrip(host, port, {"version"}))
        -> (json) {'STATUS': [{'Code': 22, 'Description': 'LUXminer 20 ...
        print(await roundtrip(host, port, "version"))
        -> (str) "{'STATUS': [{'Code': 22, 'Description': 'LUXminer 20 ..
    """
    timeout = TIMEOUT if timeout is None else timeout
    retry = RETRIES if retry is None else retry
    retry_delay = RETRIES_DELAY if retry_delay is None else retry_delay

    if not isinstance(cmd, (bytes, str)):
        cmd = json.dumps(cmd, indent=2, sort_keys=True)
        if asjson is None:
            asjson = True

    last_exception = None
    for _ in range(max(retry, 1)):
        try:
            res = await _roundtrip(host, port, cmd, timeout)
            if asjson:
                return json.loads(res)
            else:
                return res
        except (Exception, asyncio.TimeoutError) as e:
            last_exception = e
        if retry and retry_delay:
            await asyncio.sleep(retry_delay)

    if last_exception is not None:
        raise exceptions.MinerCommandTimeoutError(host, port) from last_exception


def validate_message(
    host: str,
    port: int,
    res: dict[str, Any],
    extrakey: str | None = None,
    minfields: None | int = 1,
    maxfields: None | int = 1,
) -> Any:
    # all miner message comes with a STATUS
    for key in ["STATUS", "id"]:
        if key in res:
            continue
        raise exceptions.MinerCommandMalformedMessageError(
            host, port, f"missing {key} from message STATUS", res
        )

    # no further validation here
    if not extrakey:
        return res

    if not res["STATUS"] or not res["STATUS"][0].get("STATUS") == "S":
        raise exceptions.MinerCommandMalformedMessageError(
            host, port, "no status information in message", res
        )

    # TODO if minfield is 0, there might not be extrakey
    if minfields == 0 and extrakey not in res:
        return []

    if extrakey not in res:
        if not minfields:
            return []
        raise exceptions.MinerCommandMalformedMessageError(
            host, port, f"missing {extrakey} from message", res
        )

    n = len(res[extrakey])
    msg = None

    cond = ""
    if minfields is not None and maxfields is None:
        cond = f" ({n} < {minfields})"
    elif minfields is None and maxfields is not None:
        cond = f" ({n} > {maxfields})"
    elif minfields is not None and maxfields is not None:
        if n > maxfields:
            cond = f" ({n} > {maxfields})"
        else:
            cond = f" ({n} < {minfields})"

    if (minfields is not None) and (n < minfields):
        msg = f"found too few items for '{extrakey}' {cond}"
    elif (maxfields is not None) and (n > maxfields):
        msg = f"found too many items for '{extrakey}' {cond}"
    if msg is None:
        return res
    raise exceptions.MinerCommandMalformedMessageError(host, port, msg, res)


def validate(
    res: dict[str, Any],
    extrakey: str | None = None,
    minfields: None | int = None,
    maxfields: None | int = None,
) -> Any:
    """
    Validate a message returned from a miner.

    Args:
        res: The dictionary returned from a miner.
        extrakey: if present it will try to extract it from res.
        cmd: A string representing the command to execute.
        minfields: The min length of the field extrakey.
        maxfields: The max length of the field extrakey.

    Returns:
        A dictionary containing the response from the execution of the
        command if extrakey is None, else a dict or a list of items.

    Raises:

        RuntimeError: if minfield > maxfield (internal error).
        MinerMessageMalformedError: the res dict is missing either
            'STATUS' or 'id' keys.
        MinerMessageError: the message is well formed but STATUS is not `S`.
        MinerMessageMalformedError: the res[extrakey] is not a list
        MinerMessageInvalidError: the message is missing `extarkey` or
            the len(res[extrakey]) is not valid.
    """

    if minfields is not None and maxfields is not None:
        if minfields > maxfields:
            raise RuntimeError(f"invalid arguments: {minfields=} > {maxfields=}")
    # all miner message comes with a STATUS
    for key in ["STATUS", "id"]:
        if key in res:
            continue
        raise exceptions.MinerMessageMalformedError(
            f"missing {key} from message STATUS", res
        )

    # no further validation here
    if not extrakey:
        return res

    if (status := res["STATUS"][0].get("STATUS")) != "S":
        raise exceptions.MinerMessageError(
            f"wrong status '{status}' in message (expected S)", res
        )
    # ok, when there aren't pools, the POOLS command
    # doesn't add a 'POOLS': [] item
    if extrakey not in res and not minfields:
        return None

    if extrakey not in res:
        raise exceptions.MinerMessageInvalidError(
            f"missing {extrakey} from message", res
        )

    values = res[extrakey]
    if not isinstance(values, list):
        raise exceptions.MinerMessageMalformedError(
            f"message reply doesn't contain list in '{extrakey}'", res
        )

    n = len(values)
    msg = None

    cond = ""
    if minfields is not None and maxfields is None:
        cond = f" ({n} < {minfields})"
    elif minfields is None and maxfields is not None:
        cond = f" ({n} > {maxfields})"
    elif minfields is not None and maxfields is not None:
        if n > maxfields:
            cond = f" ({n} > {maxfields})"
        else:
            cond = f" ({n} < {minfields})"

    if (minfields is not None) and (n < minfields):
        msg = f"found too few items for '{extrakey}' {cond}"
    elif (maxfields is not None) and (n > maxfields):
        msg = f"found too many items for '{extrakey}' {cond}"
    if msg is not None:
        raise exceptions.MinerMessageInvalidError(msg, res)

    if isinstance(values, list):
        ones = (len(values), minfields, maxfields)
        if ones == (1, 1, 1):
            return values[0]
    return values


@wrapped
async def logon(host: str, port: int, timeout: float | None = None) -> str:
    timeout = TIMEOUT if timeout is None else timeout
    res = await roundtrip(host, port, {"command": "logon"}, timeout=timeout)

    # when we first logon, we'll receive a token (session_id)
    #   [STATUS][SessionID]
    # on subsequent logon, we receive a
    #   [STATUS][Msg] == "Another session is active" ([STATUS][Code] 402)
    if "SESSION" not in res and res.get("STATUS", [{}])[0].get("Code") == 402:
        raise exceptions.MinerCommandSessionAlreadyActive(
            host, port, "connection active", res
        )
    sessions = validate_message(host, port, res, "SESSION", 1, 1)["SESSION"]

    session = sessions[0]

    if "SessionID" not in session:
        raise exceptions.MinerCommandSessionAlreadyActive(
            host, port, "no SessionID in data", res
        )
    return str(session["SessionID"])


@wrapped
async def logoff(
    host: str, port: int, sid: str, timeout: float | None = None
) -> dict[str, Any]:
    timeout = TIMEOUT if timeout is None else timeout
    return await roundtrip(
        host, port, {"command": "logoff", "parameter": sid}, timeout=timeout
    )


def parameters_to_list(
    parameters: str | int | float | bool | list[Any] | dict[str, Any] | None = None,
) -> list[str]:
    if isinstance(parameters, dict):
        data = []
        for key, value in parameters.items():
            if value is None:
                value = "null"
            elif value is True:
                value = "true"
            elif value is False:
                value = "false"
            data.append(f"{key}={value}")
        parameters = data
    parameters = (
        [parameters] if isinstance(parameters, (int, float, str)) else parameters
    ) or []
    parameters = [str(param) for param in parameters]
    return parameters


async def rexec(
    host: str,
    port: int,
    cmd: str,
    parameters: str | int | float | bool | list[Any] | dict[str, Any] | None = None,
    timeout: float | None = None,
    retry: int | None = None,
    retry_delay: float | None = None,
) -> dict[str, Any]:
    """
    Send a command to a host.

    Args:
        host: A string representing the host IP or a name.
        port: An integer representing the port number to connect to.
        cmd: A string representing the command to execute.
        parameters: Any additional parameters for the command.
        timeout: A float representing the maximum time in seconds to
            wait for a response before timing out.
        retry: Optional. An integer representing the number of times
            to retry the command execution in case of failure.
        retry_delay: Optional. A float representing the delay in seconds
            between each retry attempt.

    Returns:
        A dictionary containing the response from the execution of the command.

    Raises:
        Any exception that occurs during the execution of the command.

    Notes:
        If `timeout`/`retry`/`retry_delay` aren't provided (or None),
        they will default to the module level values
        (:py:data:`TIMEOUT`, :py:data:`RETRIES`, and :py:data:`RETRIES_DELAY`).

        This function will handle logon/logoff automatically.

    """

    parameters = parameters_to_list(parameters)

    timeout = TIMEOUT if timeout is None else timeout
    retry = RETRIES if retry is None else retry
    retry_delay = RETRIES_DELAY if retry_delay is None else retry_delay

    # if cmd is logon/logoff we dealt with it differently
    if cmd in {"logon", "logoff"}:
        failure = None
        for i in range(retry or 1):
            try:
                if cmd == "logon":
                    return {"sid": await logon(host, port, timeout)}
                else:
                    return await logoff(host, port, parameters[0])
            except Exception as exc:
                failure = exc
            if retry and (i < retry) and retry_delay:
                await asyncio.sleep(retry_delay)
        if isinstance(failure, Exception):
            raise failure

    failure = None
    sid = ""
    for i in range(retry + 1):
        if not api.logon_required(cmd):
            log.debug("no logon required for command '%s' on %s:%i", cmd, host, port)
            break
        try:
            sid = await logon(host, port, timeout)
            parameters = [sid, *parameters]
            log.debug("session id requested & obtained for %s:%i (%s)", host, port, sid)
            break
        except Exception as exc:
            failure = exc
        if retry and (i < retry) and retry_delay:
            await asyncio.sleep(retry_delay)

    if isinstance(failure, Exception):
        raise failure

    packet = {"command": cmd}
    if parameters:
        packet["parameter"] = ",".join(parameters)
    log.debug(
        "executing command '%s' on '%s:%i' with parameters: %s",
        cmd,
        host,
        port,
        packet.get("parameter", ""),
    )

    failure = None
    for i in range(retry + 1):
        try:
            ret = await roundtrip(host, port, packet, timeout=timeout)
            log.debug("received from %s:%s: %s", host, port, str(ret))
            if sid:
                await logoff(host, port, sid)
            return ret
        except Exception as exc:
            failure = exc
        if retry and (i < retry) and retry_delay:
            log.debug("failed attempt %i (out of %i)", i + 1, retry)
            await asyncio.sleep(retry_delay)

    if sid:
        await logoff(host, port, sid)
    if isinstance(failure, Exception):
        raise failure
    return {}


@contextlib.asynccontextmanager
async def with_atm(host, port, enabled: bool, timeout: float | None = None):
    res = await rexec(host, port, "atm", timeout=timeout)
    if not res:
        raise exceptions.MinerConnectionError(host, port, "cannot check atm")
    current = validate_message(host, port, res, "ATM")["ATM"][0]["Enabled"]
    await rexec(host, port, "atmset", {"enabled": enabled}, timeout=timeout)
    # TODO
    yield current
    await rexec(host, port, "atmset", {"enabled": current}, timeout=timeout)



from __future__ import annotations

import sys
import types
from pathlib import Path

__version__ = "0.2.5b62"
__hash__ = "1124cb2cc3a64dc0b28e26bc1a6215285a291cc4"


def get_version_info(
    modules: list[types.ModuleType] | None = None, short: bool = False
) -> dict[str, str]:
    result = {
        "py": sys.version.partition(" ")[0],
        "luxos": ", ".join(
            str(c) if str(c) else "N/A"
            for c in [__version__, __hash__[:7] if short else __hash__]
        ),
    }

    module = modules[-1] if modules else None
    if module:
        if path := getattr(module, "__file__", None):
            name = Path(path).name
        else:
            name = getattr(module, "__name__", str(module))
        if name not in {
            "luxos.py",
            "luxos_run.py",
        }:
            result[name] = getattr(module, "__version__", "N/A")

    return result


def get_version(modules: list[types.ModuleType] | None = None) -> str:
    return ", ".join(
        f"{k}[{v}]" for k, v in get_version_info(modules, short=True).items()
    )



"""Script to issue commands to miner(s)

This tool is designed to run commands to a set of miners defined on the command
line using the --range flag (either an ip address or a file prefixed with @.

Eg.

    # my-script.py
    from luxos import asyncops
    async def main(host: str, port: int):
        res = await asyncops.rexec(host, port, "version")
        return asyncops.validate(res, "VERSION", 1, 1)

    # in the cli
    $> luxos-run --range 127.0.0.1 --json my-script.py

NOTE:
  1. you can use multiple --range flags
  2. you can pass to --range flag a file (csv or yaml) using --range @file.csv
  3. ranges can specify a ip-start:ip-end
"""

from __future__ import annotations

import argparse
import asyncio
import inspect
import json
import logging
import pickle
import sys
from pathlib import Path

from luxos import misc, text, utils

from ..cli import v1 as cli

log = logging.getLogger(__name__)


def add_arguments(parser: cli.LuxosParserBase) -> None:
    cli.flags.add_arguments_new_miners_ips(parser)
    cli.flags.add_arguments_rexec(parser)
    parser.add_argument("script", type=Path, help="python script to run")
    parser.add_argument(
        "-e",
        "--entry-point",
        dest="entrypoint",
        help="script entry point",
        default="main",
    )
    parser.add_argument(
        "-t", "--teardown", dest="teardown", help="script tear down function"
    )
    parser.add_argument("-n", "--batch", type=int, help="limit parallel executions")
    parser.add_argument(
        "--list", action="store_true", help="just display the machine to run script"
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument("--json", action="store_true", help="json output")
    group.add_argument("--pickle", type=Path, help="pickle output")


def process_args(args: argparse.Namespace):
    if not args.addresses:
        args.error("need a miners flag (eg. --range/--ipfile)")
    if not args.script.exists():
        args.error(f"missings script file {args.script}")


@cli.cli(add_arguments=add_arguments, process_args=process_args)
async def main(args: argparse.Namespace):
    if args.list:
        for address in args.addresses:
            print(address)
        return

    # prepend the script dir to pypath
    log.debug("inserting %s in PYTHONPATH", args.script.parent)
    sys.path.insert(0, str(args.script.parent))

    module = misc.loadmod(args.script)

    entrypoint = getattr(module, args.entrypoint, None)
    if not entrypoint:
        args.error(f"no entry point {args.entrypoint} in {args.script}")
        return

    teardown = None
    if args.teardown == "":
        pass
    elif args.teardown:
        if not hasattr(module, args.teardown):
            args.error(f"no tear down function {args.teardown} in {args.script}")
        teardown = getattr(module, args.teardown, None)
    elif hasattr(module, "teardown"):
        teardown = getattr(module, "teardown")

    result = {}

    def callback(result):
        log.info("processed %i / %i", len(result), len(args.addresses))

    for data in await utils.launch(
        args.addresses, entrypoint, batch=args.batch, asobj=True, callback=callback
    ):
        if isinstance(data, utils.LuxosLaunchTimeoutError):
            log.warning(
                "failed connection to %s: %s\n%s",
                data.address,
                data.brief,
                text.indent(data.traceback or "", "| "),
            )
        elif isinstance(data, utils.LuxosLaunchError):
            log.warning(
                "insternal error from %s: %s\n%s",
                data.address,
                data.brief,
                text.indent(data.traceback or "", "| "),
            )
        else:
            result[data.address] = data.data

    if teardown:
        if "result" in inspect.signature(teardown).parameters:
            newresult = teardown(result)
        else:
            newresult = teardown()
        result = newresult or result

    if args.json:
        print(json.dumps(result, indent=2))
    if args.pickle:
        args.pickle.write_bytes(pickle.dumps(result))


def run():
    asyncio.run(main())


if __name__ == "__main__":
    run()



import contextlib
import sys
import types
from unittest import mock

import pytest

from luxos import version

PYVERSION = sys.version.partition(" ")[0]

VALUES = [
    # version, hash, expected
    ("", "", "N/A, N/A"),
    ("1.2.3", "", "1.2.3, N/A"),
    ("1.2.3", "abcdef", "1.2.3, abcdef"),
    ("", "abcdef", "N/A, abcdef"),
    ("", "1234567XXXX", "N/A, 1234567"),  # hash is truncated to 7chars
]


@pytest.mark.parametrize("mversion, mhash, expected", VALUES)
def test_no_extra_modules(mversion, mhash, expected):
    with contextlib.ExitStack() as stack:
        stack.enter_context(mock.patch("luxos.version.__version__", mversion))
        stack.enter_context(mock.patch("luxos.version.__hash__", mhash))
        assert version.get_version_info() == {
            "py": PYVERSION,
            "luxos": f"{mversion or 'N/A'}, {mhash or 'N/A'}",
        }
        assert version.get_version() == f"py[{PYVERSION}], luxos[{expected}]"


def test_add_module_info():
    module = types.ModuleType("xyz")
    with contextlib.ExitStack() as stack:
        stack.enter_context(mock.patch("luxos.version.__version__", "1.2.4"))
        stack.enter_context(mock.patch("luxos.version.__hash__", "a123b"))
        assert version.get_version_info([module]) == {
            "py": PYVERSION,
            "luxos": "1.2.4, a123b",
            "xyz": "N/A",
        }

    module.__name__ = "ABC"
    with contextlib.ExitStack() as stack:
        stack.enter_context(mock.patch("luxos.version.__version__", "1.2.4"))
        stack.enter_context(mock.patch("luxos.version.__hash__", "a123b"))
        assert version.get_version_info([module]) == {
            "py": PYVERSION,
            "luxos": "1.2.4, a123b",
            "ABC": "N/A",
        }

    module.__version__ = "9.9.9"
    with contextlib.ExitStack() as stack:
        stack.enter_context(mock.patch("luxos.version.__version__", "1.2.4"))
        stack.enter_context(mock.patch("luxos.version.__hash__", "a123b"))
        assert version.get_version_info([module]) == {
            "py": PYVERSION,
            "luxos": "1.2.4, a123b",
            "ABC": "9.9.9",
        }



"""cli utilities

This is the v1.cli decorator to help writing cli scripts with a consistent
interface.

A simple v1 script will always have:
* `-v/--verbose | -q/--quiet` flags to increase the logging verbosity level
* `-c/--config` to pass a config file (default to config.yaml)

A `sample.py` script with default sensible and consistent interface:

```
import argparse
import luxos.cli.v1 as cli
import logging

log = logging.getLogger(__name__)


@cli.cli()
def main(args: argparse.Namespace):
    log.debug("a debug message, need to use -v|--verbose to display it")
    log.info("an info message, you can silence it with -q|--quiet")
    log.warning("a warning!")


if __name__ == "__main__":
    main()
```


### Advanced usages

#### changing the default config file

In the `sample.py` file, just add:

```
CONFIGPATH = "my.config.file.yaml"
```

`CONFIGPATH` is part of the module level "magic" variables.

Another one is `LOGGING_CONFIG` (not recommended!):
```
LOGGING_CONFIG = {
    'level': logging.INFO,
    'format': "%(asctime)s:%(name)s:[%(levelname)s] %(message)s",
    'handlers': [
        logging.StreamHandler(),
        logging.FileHandler("LuxOS-LoadControl.log")
    ],
}
```

The `LOGGING_CONFIG` is a dictionary feed into `logging.basicConfig(**LOGGING_CONFIG)`.

#### add and process new extra arguments

in the `sample.py` file:

```
def add_arguments(parser: argparse.ArgumentParser) -> None:
    ... adds as many args needed


def process_args(args: argparse.Namespace) -> argparse.Namespace | None:
    ... you can manipulate args in place, changing its attributes in place
    ... if you return a new Namespace instance then it will be fee to main below


@cli.cli(add_arguments, process_args)
def main(args: argparse.Namespace):
    ... args is the Namespace instance returned by process_args if defined,
    ... or the default processed one of process_args is not provided

```

#### escape hatch to handle all the parsing/processing

```

@cli.cli()
def main(parser: argparse.ArgumentParser):
    ... you can use directly the parser here

```

"""

from __future__ import annotations

import argparse
import contextlib
import functools
import inspect
import logging.handlers
import sys
import time
import types
from pathlib import Path
from typing import Any, Callable

from . import flags
from .shared import ArgumentTypeBase, LuxosParserBase


class MyHandler(logging.StreamHandler):
    def emit(self, record):
        record.shortname = record.levelname[0]
        return super().emit(record)


# SPECIAL MODULE LEVEL VARIABLES
LOGGING_CONFIG = {
    "format": "%(asctime)s [%(shortname)s] %(name)s: %(message)s",
    "handlers": [
        MyHandler(),
    ],
}
CONFIGPATH = Path("config.yaml")


log = logging.getLogger(__name__)


class CliBaseError(Exception):
    pass


class AbortCliError(CliBaseError):
    pass


class AbortWrongArgumentError(CliBaseError):
    pass


def log_sys_info(modules=None):
    from luxos.version import get_version

    log.info(get_version())
    log.debug("interpreter: %s", sys.executable)


ArgumentParser = LuxosParserBase


class LuxosParser(LuxosParserBase):
    def __init__(self, modules: list[types.ModuleType], *args, **kwargs):
        from luxos.version import get_version

        super().__init__(modules, *args, **kwargs)

        # we're adding the -v|-q flags, to control the logging level
        flags.add_arguments_logging(self)

        # and a --version flag
        self.add_argument("--version", action="version", version=get_version(modules))

    def error(self, message: str):
        raise AbortWrongArgumentError(message)

    def parse_args(self, args=None, namespace=None):
        options = super().parse_args(args, namespace)

        # reserver attributes
        for reserved in [
            "modules",
            "error",
        ]:
            if not hasattr(options, reserved):
                continue
            raise RuntimeError(f"cannot add an argument with dest='{reserved}'")
        options.error = self.error
        options.modules = self.modules

        for name in dir(options):
            if isinstance(getattr(options, name), ArgumentTypeBase):
                setattr(options, name, getattr(options, name).value)

        for callback in self.callbacks:
            if not callback:
                continue
            options = callback(options) or options

        log_sys_info(self.modules)
        return options

    @classmethod
    def get_parser(cls, modules: list[types.ModuleType], **kwargs):
        class Formatter(
            argparse.RawTextHelpFormatter,
            argparse.RawDescriptionHelpFormatter,
            argparse.ArgumentDefaultsHelpFormatter,
        ):
            pass

        return cls(modules, formatter_class=Formatter, **kwargs)


@contextlib.contextmanager
def setup(
    function: Callable,
    add_arguments: (
        Callable[[LuxosParserBase], None]
        | Callable[[argparse.ArgumentParser], None]
        | None
    ) = None,
    process_args: (
        Callable[[argparse.Namespace], argparse.Namespace | None] | None
    ) = None,
):
    from luxos.version import get_version

    sig = inspect.signature(function)
    module = inspect.getmodule(function)

    if "args" in sig.parameters and "parser" in sig.parameters:
        raise RuntimeError("cannot use args and parser at the same time")

    description, _, epilog = (
        (function.__doc__ or module.__doc__ or "").strip().partition("\n")
    )
    epilog = f"{description}\n{'-'*len(description)}\n{epilog}"
    description = ""

    kwargs = {}
    modules = [
        sys.modules[__name__],
    ]
    if module:
        modules.append(module)
    parser = LuxosParser.get_parser(modules, description=description, epilog=epilog)
    if add_arguments and (callbacks := add_arguments(parser)):
        if isinstance(callbacks, list):
            parser.callbacks.extend(callbacks)
        else:
            parser.callbacks.append(callbacks)

    if "parser" in sig.parameters:
        kwargs["parser"] = parser

    t0 = time.monotonic()
    success = "completed"
    errormsg = ""
    show_timing = True
    try:
        if "parser" not in sig.parameters:
            args = parser.parse_args()
            if process_args:
                args = process_args(args) or args

            if "args" in sig.parameters:
                kwargs["args"] = args
        yield sig.bind(**kwargs)
    except AbortCliError as exc:
        show_timing = False
        if exc.args:
            print(str(exc), file=sys.stderr)
        sys.exit(2)
    except AbortWrongArgumentError as exc:
        show_timing = False
        parser.print_usage(sys.stderr)
        print(f"{parser.prog}: error: {exc.args[0]}", file=sys.stderr)
        sys.exit(2)
    except SystemExit as exc:
        show_timing = False
        sys.exit(exc.code)
    except Exception:
        log.exception("un-handled exception")
        success = f"failed ({get_version(modules)})"
    finally:
        if show_timing:
            delta = round(time.monotonic() - t0, 2)
            log.info("task %s in %.2fs", success, delta)
    if errormsg:
        parser.error(errormsg)


def cli(
    # add_arguments: Callable[[LuxosParserBase | argparse.ArgumentParser], Any]
    add_arguments: (
        Callable[[LuxosParserBase], Any]
        | Callable[[argparse.ArgumentParser], Any]
        | None
    ) = None,
    process_args: (
        Callable[[argparse.Namespace], argparse.Namespace | None] | None
    ) = None,
):
    def _cli1(function):
        module = inspect.getmodule(function)

        if inspect.iscoroutinefunction(function):

            @functools.wraps(function)
            async def _cli2(*args, **kwargs):
                with setup(function, add_arguments, process_args) as ba:
                    return await function(*ba.args, **ba.kwargs)

        else:

            @functools.wraps(function)
            def _cli2(*args, **kwargs):
                with setup(function, add_arguments, process_args) as ba:
                    return function(*ba.args, **ba.kwargs)

        _cli2.attributes = {
            "doc": function.__doc__ or module.__doc__ or "",
        }
        return _cli2

    return _cli1



import asyncio


class LuxosBaseException(Exception):
    pass


# for all broken messages coming back from miners
class MinerMessageReplyError(LuxosBaseException):
    pass


# messages missing the STATUS/id field
class MinerMessageMalformedError(MinerMessageReplyError):
    pass


# messages with a STATUS = 'E'
class MinerMessageError(MinerMessageReplyError):
    pass


# message reply format invalid
class MinerMessageInvalidError(MinerMessageReplyError):
    pass


class MinerConnectionError(LuxosBaseException):
    def __init__(self, host: str, port: int, *args, **kwargs):
        super().__init__(host, port, *args, **kwargs)
        self.address = (host, port)

    def __str__(self):
        msg = "unknown reason"
        if getattr(self, "__cause__"):
            msg = repr(self.__cause__)
        elif self.args[2:]:
            msg = str(self.args[2])
        return (
            f"<{self.address[0]}:{self.address[1]}>: {self.__class__.__name__}, "
            f"{msg}"
        )


class MinerCommandTimeoutError(MinerConnectionError, asyncio.TimeoutError):
    pass


class MinerCommandSessionAlreadyActive(MinerConnectionError):
    pass


class MinerCommandMalformedMessageError(MinerConnectionError):
    pass


class MinerCommandFailedError(MinerConnectionError):
    pass


class AddressParsingError(LuxosBaseException):
    pass



from __future__ import annotations

import importlib.resources
import json
import logging

log = logging.getLogger(__name__)


COMMANDS = json.loads((importlib.resources.files("luxos") / "api.json").read_text())


def logon_required(cmd: str, commands_list=COMMANDS) -> bool | None:
    # Check if the command requires logon to LuxOS API

    if cmd not in COMMANDS:
        log.info(
            "%s command is not supported. Try again with a different command.", cmd
        )
        return None

    return COMMANDS[cmd]["logon_required"]
